"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = _default;
exports.disconnect = disconnect;
exports.wrapIdentifier = wrapIdentifier;
exports.getQuerySelectTop = getQuerySelectTop;
exports.query = query;
exports.executeQuery = executeQuery;
exports.listTables = listTables;
exports.listViews = listViews;
exports.listRoutines = listRoutines;
exports.listTableColumns = listTableColumns;
exports.listTableTriggers = listTableTriggers;
exports.listTableIndexes = listTableIndexes;
exports.listSchemas = listSchemas;
exports.listDatabases = listDatabases;
exports.getTableReferences = getTableReferences;
exports.getTableKeys = getTableKeys;
exports.getTableCreateScript = getTableCreateScript;
exports.getViewCreateScript = getViewCreateScript;
exports.getRoutineCreateScript = getRoutineCreateScript;
exports.truncateAllTables = truncateAllTables;
exports.driverExecuteQuery = driverExecuteQuery;

var _sqlite = _interopRequireDefault(require("sqlite3"));

var _sqlQueryIdentifier = require("sql-query-identifier");

var _logger = _interopRequireDefault(require("../../logger"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function ownKeys(object, enumerableOnly) { var keys = Object.keys(object); if (Object.getOwnPropertySymbols) { var symbols = Object.getOwnPropertySymbols(object); if (enumerableOnly) symbols = symbols.filter(function (sym) { return Object.getOwnPropertyDescriptor(object, sym).enumerable; }); keys.push.apply(keys, symbols); } return keys; }

function _objectSpread(target) { for (var i = 1; i < arguments.length; i++) { var source = arguments[i] != null ? arguments[i] : {}; if (i % 2) { ownKeys(Object(source), true).forEach(function (key) { _defineProperty(target, key, source[key]); }); } else if (Object.getOwnPropertyDescriptors) { Object.defineProperties(target, Object.getOwnPropertyDescriptors(source)); } else { ownKeys(Object(source)).forEach(function (key) { Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key)); }); } } return target; }

function _defineProperty(obj, key, value) { if (key in obj) { Object.defineProperty(obj, key, { value: value, enumerable: true, configurable: true, writable: true }); } else { obj[key] = value; } return obj; }

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }

const logger = (0, _logger.default)('db:clients:sqlite');
const sqliteErrors = {
  CANCELED: 'SQLITE_INTERRUPT'
};

function _default(_x, _x2) {
  return _ref.apply(this, arguments);
}

function _ref() {
  _ref = _asyncToGenerator(function* (server, database) {
    const dbConfig = configDatabase(server, database);
    logger().debug('create driver client for sqlite3 with config %j', dbConfig);
    const conn = {
      dbConfig
    }; // light solution to test connection with with the server

    const version = (yield driverExecuteQuery(conn, {
      query: 'SELECT sqlite_version() as version'
    })).data[0].version;
    conn.versionData = {
      name: 'SQLite',
      version,
      string: `SQLite ${version}`
    };
    return {
      wrapIdentifier,
      version,
      getVersion: () => conn.versionData,
      disconnect: () => disconnect(conn),
      listTables: () => listTables(conn),
      listViews: () => listViews(conn),
      listRoutines: () => listRoutines(conn),
      listTableColumns: (db, table) => listTableColumns(conn, db, table),
      listTableTriggers: table => listTableTriggers(conn, table),
      listTableIndexes: (db, table) => listTableIndexes(conn, db, table),
      listSchemas: () => listSchemas(conn),
      getTableReferences: table => getTableReferences(conn, table),
      getTableKeys: (db, table) => getTableKeys(conn, db, table),
      query: queryText => query(conn, queryText),
      executeQuery: queryText => executeQuery(conn, queryText),
      listDatabases: () => listDatabases(conn),
      getQuerySelectTop: (table, limit) => getQuerySelectTop(conn, table, limit),
      getTableCreateScript: table => getTableCreateScript(conn, table),
      getViewCreateScript: view => getViewCreateScript(conn, view),
      getRoutineCreateScript: routine => getRoutineCreateScript(conn, routine),
      truncateAllTables: () => truncateAllTables(conn)
    };
  });
  return _ref.apply(this, arguments);
}

function disconnect() {
  // SQLite does not have connection poll. So we open and close connections
  // for every query request. This allows multiple request at same time by
  // using a different thread for each connection.
  // This may cause connection limit problem. So we may have to change this at some point.
  return Promise.resolve();
}

function wrapIdentifier(value) {
  if (value === '*') return value;
  const matched = value.match(/(.*?)(\[[0-9]\])/); // eslint-disable-line no-useless-escape

  if (matched) return wrapIdentifier(matched[1]) + matched[2];
  return `"${value.replace(/"/g, '""')}"`;
}

function getQuerySelectTop(client, table, limit) {
  return `SELECT * FROM ${wrapIdentifier(table)} LIMIT ${limit}`;
}

function query(conn, queryText) {
  let queryConnection = null;
  return {
    execute() {
      return runWithConnection(conn, /*#__PURE__*/function () {
        var _ref2 = _asyncToGenerator(function* (connection) {
          try {
            queryConnection = connection;
            const result = yield executeQuery({
              connection
            }, queryText);
            return result;
          } catch (err) {
            if (err.code === sqliteErrors.CANCELED) {
              err.sqlectronError = 'CANCELED_BY_USER';
            }

            throw err;
          }
        });

        return function (_x3) {
          return _ref2.apply(this, arguments);
        };
      }());
    },

    cancel() {
      return _asyncToGenerator(function* () {
        if (!queryConnection) {
          throw new Error('Query not ready to be canceled');
        }

        queryConnection.interrupt();
      })();
    }

  };
}

function executeQuery(_x4, _x5) {
  return _executeQuery.apply(this, arguments);
}

function _executeQuery() {
  _executeQuery = _asyncToGenerator(function* (conn, queryText) {
    const result = yield driverExecuteQuery(conn, {
      query: queryText,
      multiple: true
    });
    return result.map(parseRowQueryResult);
  });
  return _executeQuery.apply(this, arguments);
}

function listTables(_x6) {
  return _listTables.apply(this, arguments);
}

function _listTables() {
  _listTables = _asyncToGenerator(function* (conn) {
    const sql = `
    SELECT name
    FROM sqlite_master
    WHERE type='table'
    ORDER BY name
  `;

    const _yield$driverExecuteQ = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ.data;

    return data;
  });
  return _listTables.apply(this, arguments);
}

function listViews(_x7) {
  return _listViews.apply(this, arguments);
}

function _listViews() {
  _listViews = _asyncToGenerator(function* (conn) {
    const sql = `
    SELECT name
    FROM sqlite_master
    WHERE type = 'view'
  `;

    const _yield$driverExecuteQ2 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ2.data;

    return data;
  });
  return _listViews.apply(this, arguments);
}

function listRoutines() {
  return Promise.resolve([]); // DOES NOT SUPPORT IT
}

function listTableColumns(_x8, _x9, _x10) {
  return _listTableColumns.apply(this, arguments);
}

function _listTableColumns() {
  _listTableColumns = _asyncToGenerator(function* (conn, database, table) {
    const sql = `PRAGMA table_info('${table}')`;

    const _yield$driverExecuteQ3 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ3.data;

    return data.map(row => ({
      columnName: row.name,
      dataType: row.type
    }));
  });
  return _listTableColumns.apply(this, arguments);
}

function listTableTriggers(_x11, _x12) {
  return _listTableTriggers.apply(this, arguments);
}

function _listTableTriggers() {
  _listTableTriggers = _asyncToGenerator(function* (conn, table) {
    const sql = `
    SELECT name
    FROM sqlite_master
    WHERE type = 'trigger'
      AND tbl_name = '${table}'
  `;

    const _yield$driverExecuteQ4 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ4.data;

    return data.map(row => row.name);
  });
  return _listTableTriggers.apply(this, arguments);
}

function listTableIndexes(_x13, _x14, _x15) {
  return _listTableIndexes.apply(this, arguments);
}

function _listTableIndexes() {
  _listTableIndexes = _asyncToGenerator(function* (conn, database, table) {
    const sql = `PRAGMA INDEX_LIST('${table}')`;

    const _yield$driverExecuteQ5 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ5.data;

    return data.map(row => row.name);
  });
  return _listTableIndexes.apply(this, arguments);
}

function listSchemas() {
  return Promise.resolve([]); // DOES NOT SUPPORT IT
}

function listDatabases(_x16) {
  return _listDatabases.apply(this, arguments);
}

function _listDatabases() {
  _listDatabases = _asyncToGenerator(function* (conn) {
    const result = yield driverExecuteQuery(conn, {
      query: 'PRAGMA database_list;'
    });
    return result.data.map(row => row.file || ':memory:');
  });
  return _listDatabases.apply(this, arguments);
}

function getTableReferences() {
  return Promise.resolve([]); // TODO: not implemented yet
}

function getTableKeys() {
  return Promise.resolve([]); // TODO: not implemented yet
}

function getTableCreateScript(_x17, _x18) {
  return _getTableCreateScript.apply(this, arguments);
}

function _getTableCreateScript() {
  _getTableCreateScript = _asyncToGenerator(function* (conn, table) {
    const sql = `
    SELECT sql
    FROM sqlite_master
    WHERE name = '${table}';
  `;

    const _yield$driverExecuteQ6 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ6.data;

    return data.map(row => row.sql);
  });
  return _getTableCreateScript.apply(this, arguments);
}

function getViewCreateScript(_x19, _x20) {
  return _getViewCreateScript.apply(this, arguments);
}

function _getViewCreateScript() {
  _getViewCreateScript = _asyncToGenerator(function* (conn, view) {
    const sql = `
    SELECT sql
    FROM sqlite_master
    WHERE name = '${view}';
  `;

    const _yield$driverExecuteQ7 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ7.data;

    return data.map(row => row.sql);
  });
  return _getViewCreateScript.apply(this, arguments);
}

function getRoutineCreateScript() {
  return Promise.resolve([]); // DOES NOT SUPPORT IT
}

function truncateAllTables(_x21) {
  return _truncateAllTables.apply(this, arguments);
}

function _truncateAllTables() {
  _truncateAllTables = _asyncToGenerator(function* (conn) {
    yield runWithConnection(conn, /*#__PURE__*/function () {
      var _ref6 = _asyncToGenerator(function* (connection) {
        const connClient = {
          connection
        };
        const tables = yield listTables(connClient);
        const truncateAll = tables.map(table => `
      DELETE FROM ${table.name};
    `).join(''); // TODO: Check if sqlite_sequence exists then execute:
        // DELETE FROM sqlite_sequence WHERE name='${table}';

        yield driverExecuteQuery(connClient, {
          query: truncateAll
        });
      });

      return function (_x25) {
        return _ref6.apply(this, arguments);
      };
    }());
  });
  return _truncateAllTables.apply(this, arguments);
}

function configDatabase(server, database) {
  return {
    database: database.database
  };
}

function parseRowQueryResult({
  data,
  statement,
  changes
}) {
  // Fallback in case the identifier could not reconize the command
  const isSelect = Array.isArray(data);
  const rows = data || [];
  return {
    command: statement.type || isSelect && 'SELECT',
    rows,
    fields: Object.keys(rows[0] || {}).map(name => ({
      name
    })),
    rowCount: data && data.length,
    affectedRows: changes || 0
  };
}

function identifyCommands(queryText) {
  try {
    return (0, _sqlQueryIdentifier.identify)(queryText, {
      strict: false
    });
  } catch (err) {
    return [];
  }
}

function driverExecuteQuery(conn, queryArgs) {
  const runQuery = (connection, {
    executionType,
    text
  }) => new Promise((resolve, reject) => {
    const method = resolveExecutionType(executionType);
    connection[method](text, queryArgs.params, function driverExecQuery(err, data) {
      if (err) {
        return reject(err);
      }

      return resolve({
        data,
        lastID: this.lastID,
        changes: this.changes
      });
    });
  });

  const identifyStatementsRunQuery = /*#__PURE__*/function () {
    var _ref3 = _asyncToGenerator(function* (connection) {
      const statements = identifyCommands(queryArgs.query);
      const results = yield Promise.all(statements.map( /*#__PURE__*/function () {
        var _ref4 = _asyncToGenerator(function* (statement) {
          const result = yield runQuery(connection, statement);
          return _objectSpread(_objectSpread({}, result), {}, {
            statement
          });
        });

        return function (_x23) {
          return _ref4.apply(this, arguments);
        };
      }()));
      return queryArgs.multiple ? results : results[0];
    });

    return function identifyStatementsRunQuery(_x22) {
      return _ref3.apply(this, arguments);
    };
  }();

  return conn.connection ? identifyStatementsRunQuery(conn.connection) : runWithConnection(conn, identifyStatementsRunQuery);
}

function runWithConnection(conn, run) {
  return new Promise((resolve, reject) => {
    const db = new _sqlite.default.Database(conn.dbConfig.database, /*#__PURE__*/function () {
      var _ref5 = _asyncToGenerator(function* (err) {
        if (err) {
          reject(err);
          return;
        }

        try {
          db.serialize();
          const results = yield run(db);
          resolve(results);
        } catch (runErr) {
          reject(runErr);
        } finally {
          db.close();
        }
      });

      return function (_x24) {
        return _ref5.apply(this, arguments);
      };
    }());
  });
}

function resolveExecutionType(executioType) {
  switch (executioType) {
    case 'MODIFICATION':
      return 'run';

    default:
      return 'all';
  }
}