"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = _default;
exports.disconnect = disconnect;
exports.wrapIdentifier = wrapIdentifier;
exports.getQuerySelectTop = getQuerySelectTop;
exports.query = query;
exports.executeQuery = executeQuery;
exports.listTables = listTables;
exports.listViews = listViews;
exports.listRoutines = listRoutines;
exports.listTableColumns = listTableColumns;
exports.listTableTriggers = listTableTriggers;
exports.listTableIndexes = listTableIndexes;
exports.listSchemas = listSchemas;
exports.listDatabases = listDatabases;
exports.getTableReferences = getTableReferences;
exports.getTableKeys = getTableKeys;
exports.getTableCreateScript = getTableCreateScript;
exports.getViewCreateScript = getViewCreateScript;
exports.getRoutineCreateScript = getRoutineCreateScript;
exports.truncateAllTables = truncateAllTables;
exports.driverExecuteQuery = driverExecuteQuery;

var _mssql = require("mssql");

var _sqlQueryIdentifier = require("sql-query-identifier");

var _utils = require("./utils");

var _logger = _interopRequireDefault(require("../../logger"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) { try { var info = gen[key](arg); var value = info.value; } catch (error) { reject(error); return; } if (info.done) { resolve(value); } else { Promise.resolve(value).then(_next, _throw); } }

function _asyncToGenerator(fn) { return function () { var self = this, args = arguments; return new Promise(function (resolve, reject) { var gen = fn.apply(self, args); function _next(value) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value); } function _throw(err) { asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err); } _next(undefined); }); }; }

const logger = (0, _logger.default)('db:clients:sqlserver');
const mmsqlErrors = {
  CANCELED: 'ECANCEL'
};

function _default(_x, _x2) {
  return _ref.apply(this, arguments);
}

function _ref() {
  _ref = _asyncToGenerator(function* (server, database) {
    const dbConfig = configDatabase(server, database);
    logger().debug('create driver client for mmsql with config %j', dbConfig);

    if (!dbConfig.options) {
      dbConfig.options = {};
    }

    if (dbConfig.options.enableArithAbort === undefined) {
      dbConfig.options.enableArithAbort = false;
    }

    const conn = {
      dbConfig
    }; // light solution to test connection with with the server

    const version = (yield driverExecuteQuery(conn, {
      query: "SELECT @@version as 'version'"
    })).data[0].version;
    conn.versionData = {
      name: 'SQL Server',
      version: version.match(/^Microsoft SQL Server ([0-9]{4})/)[1],
      string: version
    };
    return {
      wrapIdentifier,
      version,
      getVersion: () => conn.versionData,
      disconnect: () => disconnect(conn),
      listTables: (db, filter) => listTables(conn, filter),
      listViews: filter => listViews(conn, filter),
      listRoutines: filter => listRoutines(conn, filter),
      listTableColumns: (db, table) => listTableColumns(conn, db, table),
      listTableTriggers: table => listTableTriggers(conn, table),
      listTableIndexes: (db, table) => listTableIndexes(conn, db, table),
      listSchemas: () => listSchemas(conn),
      getTableReferences: table => getTableReferences(conn, table),
      getTableKeys: (db, table) => getTableKeys(conn, db, table),
      query: queryText => query(conn, queryText),
      executeQuery: queryText => executeQuery(conn, queryText),
      listDatabases: filter => listDatabases(conn, filter),
      getQuerySelectTop: (table, limit) => getQuerySelectTop(conn, table, limit),
      getTableCreateScript: table => getTableCreateScript(conn, table),
      getViewCreateScript: view => getViewCreateScript(conn, view),
      getRoutineCreateScript: routine => getRoutineCreateScript(conn, routine),
      truncateAllTables: () => truncateAllTables(conn)
    };
  });
  return _ref.apply(this, arguments);
}

function disconnect(_x3) {
  return _disconnect.apply(this, arguments);
}

function _disconnect() {
  _disconnect = _asyncToGenerator(function* (conn) {
    const connection = yield (0, _mssql.connect)(conn.dbConfig);
    connection.close();
  });
  return _disconnect.apply(this, arguments);
}

function wrapIdentifier(value) {
  return value !== '*' ? `[${value.replace(/\[/g, '[')}]` : '*';
}

function getQuerySelectTop(client, table, limit) {
  return `SELECT TOP ${limit} * FROM ${wrapIdentifier(table)}`;
}

function query(conn, queryText) {
  let queryRequest = null;
  return {
    execute() {
      return runWithConnection(conn, /*#__PURE__*/function () {
        var _ref2 = _asyncToGenerator(function* (connection) {
          const request = connection.request();
          request.multiple = true;

          try {
            const promiseQuery = request.query(queryText);
            queryRequest = request;
            const result = yield promiseQuery;
            const data = request.multiple ? result.datasets : result.dataset;
            const affectedRows = result.affectedRows.reduce((a, b) => a + b, 0);
            const commands = identifyCommands(queryText).map(item => item.type); // Executing only non select queries will not return results.
            // So we "fake" there is at least one result.

            const results = !data.length && affectedRows ? [[]] : data;
            return results.map((_, idx) => parseRowQueryResult(results[idx], affectedRows, commands[idx]));
          } catch (err) {
            if (err.code === mmsqlErrors.CANCELED) {
              err.sqlectronError = 'CANCELED_BY_USER';
            }

            throw err;
          }
        });

        return function (_x4) {
          return _ref2.apply(this, arguments);
        };
      }());
    },

    cancel() {
      return _asyncToGenerator(function* () {
        if (!queryRequest) {
          throw new Error('Query not ready to be canceled');
        }

        queryRequest.cancel();
      })();
    }

  };
}

function executeQuery(_x5, _x6) {
  return _executeQuery.apply(this, arguments);
}

function _executeQuery() {
  _executeQuery = _asyncToGenerator(function* (conn, queryText) {
    const _yield$driverExecuteQ = yield driverExecuteQuery(conn, {
      query: queryText,
      multiple: true
    }),
          data = _yield$driverExecuteQ.data,
          result = _yield$driverExecuteQ.result;

    const commands = identifyCommands(queryText).map(item => item.type); // Executing only non select queries will not return results.
    // So we "fake" there is at least one result.

    const rowsAffected = result.rowsAffected.reduce((a, b) => a + b, 0);
    const results = !data.length && rowsAffected ? [[]] : data;
    return results.map((_, idx) => parseRowQueryResult(results[idx], rowsAffected, commands[idx]));
  });
  return _executeQuery.apply(this, arguments);
}

function getSchema(_x7) {
  return _getSchema.apply(this, arguments);
}

function _getSchema() {
  _getSchema = _asyncToGenerator(function* (conn) {
    const sql = 'SELECT schema_name() AS \'schema\'';

    const _yield$driverExecuteQ2 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ2.data;

    return data[0].schema;
  });
  return _getSchema.apply(this, arguments);
}

function listTables(_x8, _x9) {
  return _listTables.apply(this, arguments);
}

function _listTables() {
  _listTables = _asyncToGenerator(function* (conn, filter) {
    const schemaFilter = (0, _utils.buildSchemaFilter)(filter, 'table_schema');
    const sql = `
    SELECT
      table_schema,
      table_name
    FROM INFORMATION_SCHEMA.TABLES
    WHERE table_type NOT LIKE '%VIEW%'
    ${schemaFilter ? `AND ${schemaFilter}` : ''}
    ORDER BY table_schema, table_name
  `;

    const _yield$driverExecuteQ3 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ3.data;

    return data.map(item => ({
      schema: item.table_schema,
      name: item.table_name
    }));
  });
  return _listTables.apply(this, arguments);
}

function listViews(_x10, _x11) {
  return _listViews.apply(this, arguments);
}

function _listViews() {
  _listViews = _asyncToGenerator(function* (conn, filter) {
    const schemaFilter = (0, _utils.buildSchemaFilter)(filter, 'table_schema');
    const sql = `
    SELECT
      table_schema,
      table_name
    FROM INFORMATION_SCHEMA.VIEWS
    ${schemaFilter ? `WHERE ${schemaFilter}` : ''}
    ORDER BY table_schema, table_name
  `;

    const _yield$driverExecuteQ4 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ4.data;

    return data.map(item => ({
      schema: item.table_schema,
      name: item.table_name
    }));
  });
  return _listViews.apply(this, arguments);
}

function listRoutines(_x12, _x13) {
  return _listRoutines.apply(this, arguments);
}

function _listRoutines() {
  _listRoutines = _asyncToGenerator(function* (conn, filter) {
    const schemaFilter = (0, _utils.buildSchemaFilter)(filter, 'routine_schema');
    const sql = `
    SELECT
      routine_schema,
      routine_name,
      routine_type
    FROM INFORMATION_SCHEMA.ROUTINES
    ${schemaFilter ? `WHERE ${schemaFilter}` : ''}
    GROUP BY routine_schema, routine_name, routine_type
    ORDER BY routine_schema, routine_name
  `;

    const _yield$driverExecuteQ5 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ5.data;

    return data.map(row => ({
      schema: row.routine_schema,
      routineName: row.routine_name,
      routineType: row.routine_type
    }));
  });
  return _listRoutines.apply(this, arguments);
}

function listTableColumns(_x14, _x15, _x16) {
  return _listTableColumns.apply(this, arguments);
}

function _listTableColumns() {
  _listTableColumns = _asyncToGenerator(function* (conn, database, table) {
    const sql = `
    SELECT column_name, data_type
    FROM INFORMATION_SCHEMA.COLUMNS
    WHERE table_name = '${table}'
    ORDER BY ordinal_position
  `;

    const _yield$driverExecuteQ6 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ6.data;

    return data.map(row => ({
      columnName: row.column_name,
      dataType: row.data_type
    }));
  });
  return _listTableColumns.apply(this, arguments);
}

function listTableTriggers(_x17, _x18) {
  return _listTableTriggers.apply(this, arguments);
}

function _listTableTriggers() {
  _listTableTriggers = _asyncToGenerator(function* (conn, table) {
    // SQL Server does not have information_schema for triggers, so other way around
    // is using sp_helptrigger stored procedure to fetch triggers related to table
    const sql = `EXEC sp_helptrigger ${wrapIdentifier(table)}`;

    const _yield$driverExecuteQ7 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ7.data;

    return data.map(row => row.trigger_name);
  });
  return _listTableTriggers.apply(this, arguments);
}

function listTableIndexes(_x19, _x20, _x21) {
  return _listTableIndexes.apply(this, arguments);
}

function _listTableIndexes() {
  _listTableIndexes = _asyncToGenerator(function* (conn, database, table) {
    // SQL Server does not have information_schema for indexes, so other way around
    // is using sp_helpindex stored procedure to fetch indexes related to table
    const sql = `EXEC sp_helpindex ${wrapIdentifier(table)}`;

    const _yield$driverExecuteQ8 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ8.data;

    return data.map(row => row.index_name);
  });
  return _listTableIndexes.apply(this, arguments);
}

function listSchemas(_x22, _x23) {
  return _listSchemas.apply(this, arguments);
}

function _listSchemas() {
  _listSchemas = _asyncToGenerator(function* (conn, filter) {
    const schemaFilter = (0, _utils.buildSchemaFilter)(filter);
    const sql = `
    SELECT schema_name
    FROM INFORMATION_SCHEMA.SCHEMATA
    ${schemaFilter ? `WHERE ${schemaFilter}` : ''}
    ORDER BY schema_name
  `;

    const _yield$driverExecuteQ9 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ9.data;

    return data.map(row => row.schema_name);
  });
  return _listSchemas.apply(this, arguments);
}

function listDatabases(_x24, _x25) {
  return _listDatabases.apply(this, arguments);
}

function _listDatabases() {
  _listDatabases = _asyncToGenerator(function* (conn, filter) {
    const databaseFilter = (0, _utils.buildDatabseFilter)(filter, 'name');
    const sql = `
    SELECT name
    FROM sys.databases
    ${databaseFilter ? `AND ${databaseFilter}` : ''}
    ORDER BY name
  `;

    const _yield$driverExecuteQ10 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ10.data;

    return data.map(row => row.name);
  });
  return _listDatabases.apply(this, arguments);
}

function getTableReferences(_x26, _x27) {
  return _getTableReferences.apply(this, arguments);
}

function _getTableReferences() {
  _getTableReferences = _asyncToGenerator(function* (conn, table) {
    const sql = `
    SELECT OBJECT_NAME(referenced_object_id) referenced_table_name
    FROM sys.foreign_keys
    WHERE parent_object_id = OBJECT_ID('${table}')
  `;

    const _yield$driverExecuteQ11 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ11.data;

    return data.map(row => row.referenced_table_name);
  });
  return _getTableReferences.apply(this, arguments);
}

function getTableKeys(_x28, _x29, _x30) {
  return _getTableKeys.apply(this, arguments);
}

function _getTableKeys() {
  _getTableKeys = _asyncToGenerator(function* (conn, database, table) {
    const sql = `
    SELECT
      tc.constraint_name,
      kcu.column_name,
      CASE WHEN tc.constraint_type LIKE '%FOREIGN%' THEN OBJECT_NAME(sfk.referenced_object_id)
      ELSE NULL
      END AS referenced_table_name,
      tc.constraint_type
    FROM INFORMATION_SCHEMA.TABLE_CONSTRAINTS AS tc
    JOIN INFORMATION_SCHEMA.KEY_COLUMN_USAGE AS kcu
      ON tc.constraint_name = kcu.constraint_name
    JOIN sys.foreign_keys as sfk
      ON sfk.parent_object_id = OBJECT_ID(tc.table_name)
    WHERE tc.table_name = '${table}'
    AND tc.constraint_type IN ('PRIMARY KEY', 'FOREIGN KEY')
  `;

    const _yield$driverExecuteQ12 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ12.data;

    return data.map(row => ({
      constraintName: row.constraint_name,
      columnName: row.column_name,
      referencedTable: row.referenced_table_name,
      keyType: row.constraint_type
    }));
  });
  return _getTableKeys.apply(this, arguments);
}

function getTableCreateScript(_x31, _x32) {
  return _getTableCreateScript.apply(this, arguments);
}

function _getTableCreateScript() {
  _getTableCreateScript = _asyncToGenerator(function* (conn, table) {
    // Reference http://stackoverflow.com/a/317864
    const sql = `
    SELECT  ('CREATE TABLE ' + so.name + ' (' +
      CHAR(13)+CHAR(10) + REPLACE(o.list, '&#x0D;', CHAR(13)) +
      ')' + CHAR(13)+CHAR(10) +
      CASE WHEN tc.constraint_name IS NULL THEN ''
           ELSE + CHAR(13)+CHAR(10) + 'ALTER TABLE ' + so.Name +
           ' ADD CONSTRAINT ' + tc.constraint_name  +
           ' PRIMARY KEY ' + '(' + LEFT(j.list, Len(j.list)-1) + ')'
      END) AS createtable
    FROM sysobjects so
    CROSS APPLY
      (SELECT
        '  ' + column_name + ' ' +
        data_type +
        CASE data_type
            WHEN 'sql_variant' THEN ''
            WHEN 'text' THEN ''
            WHEN 'ntext' THEN ''
            WHEN 'xml' THEN ''
            WHEN 'decimal' THEN '(' + cast(numeric_precision AS varchar) + ', '
                  + cast(numeric_scale AS varchar) + ')'
            ELSE coalesce('('+ CASE WHEN character_maximum_length = -1
                  THEN 'MAX'
                  ELSE cast(character_maximum_length AS varchar)
                END + ')','')
          END + ' ' +
          CASE WHEN EXISTS (
            SELECT id FROM syscolumns
            WHERE object_name(id)=so.name
            AND name=column_name
            AND columnproperty(id,name,'IsIdentity') = 1
          ) THEN
            'IDENTITY(' +
            cast(ident_seed(so.name) AS varchar) + ',' +
            cast(ident_incr(so.name) AS varchar) + ')'
          ELSE ''
          END + ' ' +
           (CASE WHEN UPPER(IS_NULLABLE) = 'NO'
                 THEN 'NOT '
                 ELSE ''
          END ) + 'NULL' +
          CASE WHEN INFORMATION_SCHEMA.COLUMNS.column_default IS NOT NULL
               THEN 'DEFAULT '+ INFORMATION_SCHEMA.COLUMNS.column_default
               ELSE ''
          END + ',' + CHAR(13)+CHAR(10)
       FROM INFORMATION_SCHEMA.COLUMNS WHERE table_name = so.name
       ORDER BY ordinal_position
       FOR XML PATH('')
    ) o (list)
    LEFT JOIN INFORMATION_SCHEMA.TABLE_CONSTRAINTS tc
    ON  tc.table_name       = so.name
    AND tc.constraint_type  = 'PRIMARY KEY'
    CROSS APPLY
        (SELECT column_name + ', '
         FROM   INFORMATION_SCHEMA.KEY_COLUMN_USAGE kcu
         WHERE  kcu.constraint_name = tc.constraint_name
         ORDER BY ordinal_position
         FOR XML PATH('')
        ) j (list)
    WHERE   xtype = 'U'
    AND name    NOT IN ('dtproperties')
    AND so.name = '${table}'
  `;

    const _yield$driverExecuteQ13 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ13.data;

    return data.map(row => row.createtable);
  });
  return _getTableCreateScript.apply(this, arguments);
}

function getViewCreateScript(_x33, _x34) {
  return _getViewCreateScript.apply(this, arguments);
}

function _getViewCreateScript() {
  _getViewCreateScript = _asyncToGenerator(function* (conn, view) {
    const sql = `SELECT OBJECT_DEFINITION (OBJECT_ID('${view}')) AS ViewDefinition;`;

    const _yield$driverExecuteQ14 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ14.data;

    return data.map(row => row.ViewDefinition);
  });
  return _getViewCreateScript.apply(this, arguments);
}

function getRoutineCreateScript(_x35, _x36) {
  return _getRoutineCreateScript.apply(this, arguments);
}

function _getRoutineCreateScript() {
  _getRoutineCreateScript = _asyncToGenerator(function* (conn, routine) {
    const sql = `
    SELECT routine_definition
    FROM INFORMATION_SCHEMA.ROUTINES
    WHERE routine_name = '${routine}'
  `;

    const _yield$driverExecuteQ15 = yield driverExecuteQuery(conn, {
      query: sql
    }),
          data = _yield$driverExecuteQ15.data;

    return data.map(row => row.routine_definition);
  });
  return _getRoutineCreateScript.apply(this, arguments);
}

function truncateAllTables(_x37) {
  return _truncateAllTables.apply(this, arguments);
}

function _truncateAllTables() {
  _truncateAllTables = _asyncToGenerator(function* (conn) {
    yield runWithConnection(conn, /*#__PURE__*/function () {
      var _ref3 = _asyncToGenerator(function* (connection) {
        const connClient = {
          connection
        };
        const schema = yield getSchema(connClient);
        const sql = `
      SELECT table_name
      FROM INFORMATION_SCHEMA.TABLES
      WHERE table_schema = '${schema}'
      AND table_type NOT LIKE '%VIEW%'
    `;

        const _yield$driverExecuteQ16 = yield driverExecuteQuery(connClient, {
          query: sql
        }),
              data = _yield$driverExecuteQ16.data;

        const disableForeignKeys = data.map(row => `
      ALTER TABLE ${wrapIdentifier(schema)}.${wrapIdentifier(row.table_name)} NOCHECK CONSTRAINT all;
    `).join('');
        const truncateAll = data.map(row => `
      DELETE FROM ${wrapIdentifier(schema)}.${wrapIdentifier(row.table_name)};
      DBCC CHECKIDENT ('${schema}.${row.table_name}', RESEED, 0);
    `).join('');
        const enableForeignKeys = data.map(row => `
      ALTER TABLE ${wrapIdentifier(schema)}.${wrapIdentifier(row.table_name)} WITH CHECK CHECK CONSTRAINT all;
    `).join('');
        yield driverExecuteQuery(connClient, {
          query: disableForeignKeys + truncateAll + enableForeignKeys,
          multiple: true
        });
      });

      return function (_x42) {
        return _ref3.apply(this, arguments);
      };
    }());
  });
  return _truncateAllTables.apply(this, arguments);
}

function configDatabase(server, database) {
  const config = {
    user: server.config.user,
    password: server.config.password,
    server: server.config.host,
    database: database.database,
    port: server.config.port,
    requestTimeout: Infinity,
    appName: server.config.applicationName || 'sqlectron',
    domain: server.config.domain,
    pool: {
      max: 5
    },
    options: {
      encrypt: server.config.ssl
    }
  };

  if (server.sshTunnel) {
    config.server = server.config.localHost;
    config.port = server.config.localPort;
  }

  return config;
}

function parseRowQueryResult(data, rowsAffected, command) {
  // Fallback in case the identifier could not reconize the command
  const isSelect = !!(data.length || !rowsAffected);
  return {
    command: command || isSelect && 'SELECT',
    rows: data,
    fields: Object.keys(data[0] || {}).map(name => ({
      name
    })),
    rowCount: data.length,
    affectedRows: rowsAffected
  };
}

function identifyCommands(queryText) {
  try {
    return (0, _sqlQueryIdentifier.identify)(queryText);
  } catch (err) {
    return [];
  }
}

function driverExecuteQuery(_x38, _x39) {
  return _driverExecuteQuery.apply(this, arguments);
}

function _driverExecuteQuery() {
  _driverExecuteQuery = _asyncToGenerator(function* (conn, queryArgs) {
    const runQuery = /*#__PURE__*/function () {
      var _ref4 = _asyncToGenerator(function* (connection) {
        const request = connection.request();

        if (queryArgs.multiple) {
          request.multiple = true;
        }

        const result = yield request.query(queryArgs.query);
        return {
          request,
          data: request.multiple ? result.recordsets : result.recordset,
          result
        };
      });

      return function runQuery(_x43) {
        return _ref4.apply(this, arguments);
      };
    }();

    return conn.connection ? runQuery(conn.connection) : runWithConnection(conn, runQuery);
  });
  return _driverExecuteQuery.apply(this, arguments);
}

function runWithConnection(_x40, _x41) {
  return _runWithConnection.apply(this, arguments);
}

function _runWithConnection() {
  _runWithConnection = _asyncToGenerator(function* (conn, run) {
    const connection = yield (0, _mssql.connect)(conn.dbConfig);
    return run(connection);
  });
  return _runWithConnection.apply(this, arguments);
}